;; -*- lisp -*-

(in-package :it.bese.ucw)

;;;; ** STANDARD-SESSION-FRAME

(defmethod make-new-action ((frame standard-session-frame) (lambda function))
  (insert-with-new-key (frame.actions frame) +action-id-length+ lambda))

(defmethod find-action ((f standard-session-frame) (action-id string))
  (gethash action-id (frame.actions f)))

(defun register-action (action-lambda)
  (make-new-action (context.current-frame *context*) action-lambda))

(defun make-new-callback (lambda &key (frame (context.current-frame *context*))
                                 (name nil))
  "Registers a new callback.

LAMBDA when the request arrives LAMBDA will be passed the
value (a string) associated with the input. If NAME is not
provided, or NIL, a random name will be generated.

Returns NAME, or the random name if NAME was NIL."
  (let ((effective-name (or name
                            (new-random-key (frame.callbacks frame)
                                            +action-id-length+))))
    (make-new-callback-using-class frame lambda effective-name)
    effective-name))

(defmethod make-new-callback-using-class ((f standard-session-frame)
                                          lambda &optional (name nil))
  "Returns the name for a query param which, when passed in a
  action generated by F will call LAMBDA passing it the value of
  the param."
  (let ((effective-name (or name
                            (new-random-key (frame.callbacks f)
                                            +action-id-length+))))
    (setf (gethash effective-name (frame.callbacks f)) lambda)
    effective-name))

(defmethod call-callback ((f standard-session-frame) (callback-id string) value)
  (aif (gethash callback-id (frame.callbacks f))
       (with-action-error-handler ()
         (funcall it value))
       nil))

(defun register-callback (callback)
  (make-new-callback callback))

(defmethod make-next-frame ((f standard-session-frame) new-id)
  (make-instance 'standard-session-frame
                 :backtracks (clone-backtracks (frame.backtracks f))
                 :window-component (frame.window-component f)
                 :id new-id))

(defmethod make-next-frame ((f null) new-id)
  (make-instance 'standard-session-frame :id new-id))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Servicing session frame

(defmethod service ((frame standard-session-frame) (context request-context))
  "Service a request for this session frame.

First reinstate all the backtracked places associated with this
frame, then call all the param callbacks and finally call the
action.

If we find an action-id in the request we try and find the
associated action and call it. Should the action-id not
correspond we warn and redisplay current frame if it was found
or pretend it's an entry-point request otherwise."
  (labels ((service-as-action ()
             (reinstate-backtracked frame)
             (call-callbacks context)
             (call-action (find-action (context.current-frame context)
                                       (find-action-id context))))
           
           (service-as-entry-point ()
             (if-bind entry-point
                      (find-entry-point (context.application context) context)
                 (call-action (entry-point.action entry-point))
                 (error 'inexistent-entry-point)))

           (call-action (action-function)
             (make-new-frame (context.session context))
             (restart-case
                 (with-action-error-handler () 
                   (funcall action-function))             
               (retry ()
                 :report "Clearout the response and retry calling the action."
                 (setf *context* (clear-context *context*))
                 (with-yaclml-stream (html-stream (context.response *context*))
                   (return-from call-action (call-action action-function)))))))
    
    (catch 'abort-action
      (if (find-action-id context)
          (if (find-action (context.current-frame context)
                           (find-action-id context))
              (service-as-action)
              (progn
                (ucw.rerl.session-frame.warn "No action with id ~S found."
                                             (find-action-id context))
                (unless (frame.window-component frame) ; new frame
                  (service-as-entry-point))))
          (service-as-entry-point)))

    (ucw.rerl.dribble "Done handling action.")
    
    (labels ((render-loop ()
               (restart-case
                   (progn
                     (catch 'abort-action
                       (with-action-error-handler ()
                         (render (frame.window-component (context.current-frame *context*))))
                       (return-from render-loop t))
                     ;; if we get here it's because a handler for the
                     ;; render threw the abort-action tag
                     (clear-response (context.response *context*))
                     ;; NB: clear-response changes the html stream,
                     ;; but *yaclml-stream* is stil bound to the old
                     ;; value.
                     (with-yaclml-stream (html-stream (context.response *context*))
                       (render-loop)))
                 (retry ()
                   :report "Retry calling RENDER."
                   (return-from render-loop (render-loop))))))
      (render-loop))
    (ucw.rerl.dribble "Done rendering.")))

(defmethod service :after ((frame standard-session-frame) (context request-context))
  (save-backtracked (context.current-frame context)))

;; Copyright (c) 2003-2005 Edward Marco Baringer
;; All rights reserved. 
;; 
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;; 
;;  - Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 
;;  - Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;; 
;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;    of its contributors may be used to endorse or promote products
;;    derived from this software without specific prior written permission.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
